"""Schema runtime classes."""

from . import exceptions
from . import framework


class Schema(object):
  def validate(self, value):
    raise NotImplementedError()

  def get_subschema(self, key):
    raise NotImplementedError()

  def __ne__(self, other):
    return not (self == other)


class AnySchema(Schema):
  def validate(self, value):
    pass

  def get_subschema(self, key):
    return self

  def __repr__(self):
    return 'AnySchema()'

  def __eq__(self, other):
    return isinstance(other, AnySchema)


any_schema = AnySchema()  # Object reuse through immutable singleton


class AndSchema(Schema):
  def __init__(self, one, two):
    self.one = one
    self.two = two

  def validate(self, value):
    self.one.validate(value)
    self.two.validate(value)

  def get_subschema(self, key):
    return AndSchema.make(self.one.get_subschema(key), self.two.get_subschema(key))

  def __repr__(self):
    return 'AndSchema(%r, %r)' % (self.one, self.two)

  def __eq__(self, other):
    return isinstance(other, AndSchema) and self.one == other.one and self.two == other.two

  @staticmethod
  def make(one, two):
    if one == any_schema: return two
    if two == any_schema: return one
    return AndSchema(one, two) if one != two else one


class ScalarSchema(Schema):
  def __init__(self, typename):
    self.typename = typename

  def validate(self, value):
    SCALAR_TYPES[self.typename](self, value)

  def get_subschema(self, key):
    raise RuntimeError('Cannot get subschema from ScalarSchema')

  def _validate_string(self, value):
    if not framework.is_str(value):
      raise exceptions.SchemaError('%r should be a string' % value)

  def _validate_int(self, value):
    if not isinstance(value, int):
      raise exceptions.SchemaError('%r should be an int' % value)

  def _validate_bool(self, value):
    if not isinstance(value, bool):
      raise exceptions.SchemaError('%r should be a bool' % value)

  def _validate_null(self, value):
    if value is not None:
      raise exceptions.SchemaError('%r should be a null' % value)

  def _validate_float(self, value):
    if not isinstance(value, float):
      raise exceptions.SchemaError('%r should be a float' % value)

  def __repr__(self):
    return 'ScalarSchema(%r)' % (self.typename)

  def __eq__(self, other):
    return isinstance(other, ScalarSchema) and self.typename == other.typename


class ListSchema(Schema):
  def __init__(self, element_schema):
    self.element_schema = element_schema

    if not isinstance(element_schema, Schema):
      raise ValueError('Expecting Schema instance in List, got %r' % element_schema)

  def validate(self, value):
    if not framework.is_list(value):
      raise exceptions.SchemaError('%r should be a list' % value)

    for i, x in enumerate(value):
      try:
        self.element_schema.validate(x)
      except Exception as e:
        raise exceptions.SchemaError('While validating element %d of %r:\n%s' % (i + 1, x, e))

  def get_subschema(self, key):
    raise RuntimeError('Cannot get subschema from ListSchema')

  def __repr__(self):
    return 'ListSchema(%r)' % (self.element_schema)

  def __eq__(self, other):
    return isinstance(other, ListSchema) and self.element_schema == other.element_schema


class TupleSchema(Schema):
  def __init__(self, field_schemas, required_fields):
    self.field_schemas = field_schemas
    self.required_fields = required_fields

    # I would have loved to do parameter validation here, but that means we can't do recursive tuple
    # schemas (i.e., X = { x : X }). So unfortunately we'll have to defer parameter validation until
    # access. Anyway, this was only used to catch programming errors :).

    # self._validate_params()

  def _validate_params(self):
    for v in self.field_schemas.values():
      if not isinstance(v, Schema):
        raise ValueError('subschemas should be instances of Schema, got %r' % v)

  def validate(self, value):
    self._validate_params()

    if not framework.is_tuple(value):
      raise exceptions.SchemaError('%r should be a tuple' % value)

    key_check = getattr(value, 'is_bound', value.__contains__)

    for k in self.required_fields:
      if not key_check(k):
        raise exceptions.SchemaError('Tuple is missing required key %r in %r' % (k, value))

    # Attach schema to tuple, so it can verify subfields
    getattr(value, 'attach_schema', nop)(self)

    # Don't validate the subfields here -- we'll do that lazily when
    # we access the subfields.

  def get_subschema(self, key):
    # The values in field_schemas are unevaluated specs
    return self.field_schemas.get(key, any_schema)

  def __repr__(self):
    return 'TupleSchema(%r, %r)' % (self.field_schemas, self.required_fields)

  def __eq__(self, other):
    return (isinstance(other, TupleSchema) and self.required_fields == other.required_fields
        and self.field_schemas == other.field_schemas)


SCALAR_TYPES = {'string': ScalarSchema._validate_string,
                'int': ScalarSchema._validate_int,
                'bool': ScalarSchema._validate_bool,
                'null': ScalarSchema._validate_null,
                'float': ScalarSchema._validate_float}


def from_spec(spec):
  """Return a schema object from a spec.

  A spec is either a string for a scalar type, or a list of 0 or 1 specs,
  or a dictionary with two elements: {'fields': { ... }, required: [...]}.
  """
  if spec == '':
    return any_schema

  if framework.is_str(spec):
    # Scalar type
    if spec not in SCALAR_TYPES:
      raise exceptions.SchemaError('Not a valid schema type: %r' % spec)
    return ScalarSchema(spec)

  if framework.is_list(spec):
    return ListSchema(spec[0] if len(spec) else any_schema)

  if framework.is_tuple(spec):
    return TupleSchema(spec.get('fields', {}), spec.get('required', []))

  raise exceptions.SchemaError('Not valid schema spec; %r' % spec)


builtin_schemas = {k: ScalarSchema(k) for k in SCALAR_TYPES.keys()}


def nop(x): pass


def validate(obj, schema):
  """Validate an object according to its own AND an externally imposed schema."""
  # Validate returned object according to its own schema
  if hasattr(obj, 'tuple_schema'):
    obj.tuple_schema.validate(obj)
  # Validate object according to externally imposed schema
  if schema:
    schema.validate(obj)
  return obj



"""GCL standard library functions."""

import math
import itertools
import functools

from os import path

from . import framework


def eager(x):
  """Force eager evaluation of a Thunk, and turn a Tuple into a dict eagerly.

  This forces that there are no unbound variables at parsing time (as opposed
  to later when the variables may be accessed).

  Only eagerly evaluates one level.
  """
  if not hasattr(x, 'items'):
    # The Thunk has been evaluated already, so that was easy :)
    return x

  return dict(x.items())


class StringInterpolationProxy(object):
  """Lazy stringification object from a tuple.

  Also support period-based element access, cuz it looks good.
  """
  def __init__(self, tup, key):
    self.tup = tup
    self.key = key

  def __getattr__(self, key):
    v = self.tup[self.key]
    # If this is a tuple, return another lazy proxy
    if isinstance(v, framework.TupleLike):
      return StringInterpolationProxy(v, key)
    return v[key]

  def __str__(self):
    return str(self.tup[self.key])


def fmt(str, args=None, env=None):
  """String interpolation.

  Normally, we'd just call str.format(**args), but we only want to evaluate
  values from the tuple which are actually used in the string interpolation,
  so we use proxy objects.

  If no args are given, we're able to take the current environment.
  """
  args = args or env
  proxies = {k: StringInterpolationProxy(args, k) for k in args.keys()}
  return str.format(**proxies)


def str_join(lst, sep=' '):
  """Behaves like string.join from Python 2."""
  return sep.join(str(x) for x in lst)


def compose_all(tups):
  """Compose all given tuples together."""
  from . import ast  # I weep for humanity
  return functools.reduce(lambda x, y: x.compose(y), map(ast.make_tuple, tups), ast.make_tuple({}))


def split(string, sep=' '):
  """Splits a string."""
  return string.split(sep)


def has_key(tup, key):
  """Return whether a given tuple has a key and the key is bound."""
  if isinstance(tup, framework.TupleLike):
    return tup.is_bound(key)
  if isinstance(tup, dict):
    return key in tup
  if isinstance(tup, list):
    if not isinstance(key, int):
      raise ValueError('Key must be integer when checking list index')
    return key < len(tup)
  raise ValueError('Not a tuple-like object: %r' % tup)


def flatten(list_of_lists):
  """Flatten a list of lists."""
  ret = []
  for lst in list_of_lists:
    if not isinstance(lst, list):
      raise ValueError('%r is not a list' % lst)
    ret.extend(lst)
  return ret


builtin_functions = {
    'eager': eager,
    'path_join': path.join,
    'join': str_join,
    'fmt': framework.EnvironmentFunction(fmt),
    'sum': sum,
    'compose_all': compose_all,
    'sorted': sorted,
    'split': split,
    'has': has_key,
    'flatten': flatten,
    'sqrt': math.sqrt,  # Only for the Pythagoras demo :)
    'len': len
    }

# Binary operators, by precedence level
binary_operators_before_juxtaposition = [
    {
      '*': lambda x, y: x * y,
      '/': lambda x, y: x / y,
      '%': lambda x, y: x % y,
    }, {
      '+': lambda x, y: x + y,
      '-': lambda x, y: x - y,
    }]
binary_operators_after_juxtaposition = [
    {
      '==': lambda x, y: x == y,
      '!=': lambda x, y: x != y,
      '<': lambda x, y: x < y,
      '<=': lambda x, y: x <= y,
      '>': lambda x, y: x > y,
      '>=': lambda x, y: x >= y,
    }, {
      # These take thunks
      'and': framework.LazyFunction(lambda x, y: x() and y()),
      'or': framework.LazyFunction(lambda x, y: x() or y()),
    }]
all_binary_operators = {
    k: v
    for os in binary_operators_before_juxtaposition + binary_operators_after_juxtaposition
    for k, v in os.items()}


unary_operators = {
    '-': lambda x: -x,
    'not': lambda x: not x,
    }




"""General runtime classes.

(Implementations of the protocols in gcl.framework).
"""
import collections
import functools
from os import path

from . import framework
from . import schema
from . import exceptions


class Tuple(framework.TupleLike):
  """Bound tuple, with lazy evaluation.

  Contains real values or Thunks. Thunks will be evaluated upon request, but
  not before.

  The parent_env is the environment in which we do lookups for values that are
  not in this Tuple (the lexically enclosing scope).
  """
  def __init__(self, tuplenode, parent_env, dict2tuple):
    self.ident = framework.obj_ident()
    self.__tuplenode = tuplenode
    self.__parent_env = parent_env
    # This mapping is for backwards compatibility. In principle, the TupleNode owns the members
    self.__items = {m.name: m.value for m in tuplenode.members}
    self.__env_cache = framework.Cache()  # Env cache so eval caching works more effectively
    self.tuple_schema = schema.AnySchema()

    # This function is only here to break a cyclic dependency on the 'ast' module
    self.dict2tuple = dict2tuple

  def dict(self):
    return self.__items

  def get(self, key, default=None):
    if key in self:
      return self[key]
    return default

  def __getitem__(self, key):
    if type(key) == int:
      raise ValueError('Trying to access tuple as a list')

    x = self.get_no_validate(key)

    return schema.validate(x, self.tuple_schema.get_subschema(key))

  def get_no_validate(self, key):
    """Return an item without validating the schema."""
    x, env = self.get_thunk_env(key)

    # Check if this is a Thunk that needs to be lazily evaluated before we
    # return it.
    if isinstance(x, framework.Thunk):
      x = framework.eval(x, env)

    return x

  def __contains__(self, key):
    return key in self.__items

  def env(self, current_scope):
    """Return an environment that will look up in current_scope for keys in
    this tuple, and the parent env otherwise.
    """
    return self.__env_cache.get(
            current_scope.ident,
            framework.Environment, current_scope, self.__parent_env, names=self.keys())

  def keys(self):
    return self.__items.keys()

  @property
  def tuples(self):
    return [self]

  def items(self):
    return list(self.iteritems())

  def iteritems(self):
    for k in self.keys():
      yield k, self[k]

  def get_thunk(self, k):
    return self.get_thunk_env(k)[0]

  def get_thunk_env(self, k):
    if k not in self.__items:
      raise exceptions.EvaluationError('Unknown key: %r in tuple %r' % (k, self))
    x = self.__items[k]

    env = self.env(self)

    # Bind this to the tuple's parent environment
    if isinstance(x, framework.BindableThunk):
      return x.bind(self.__parent_env), env
    return x, env

  def _render(self, key):
    if key in self:
      return '%s = %r' % (key, self.get_thunk(key))
    else:
      return '%s' % key

  def is_bound(self, name):
    return name in self and not self.get_thunk(name).is_unbound()

  def compose(self, tup):
    if not isinstance(tup, Tuple):
      tup = self.dict2tuple(tup)

    composite = CompositeTuple(self.tuples + tup.tuples, self.dict2tuple)
    composite.attach_schema(self.tuple_schema)
    composite.attach_schema(tup.tuple_schema)
    return composite

  def attach_schema(self, schem):
    """Add a tuple schema to this object (externally imposed)"""
    self.tuple_schema = schema.AndSchema.make(self.tuple_schema, schem)

  def get_schema_spec(self, key):
    """Return the evaluated schema expression from a subkey."""
    member_node = self.__tuplenode.member.get(key, None)
    if not member_node:
      return schema.AnySchema()

    s = framework.eval(member_node.member_schema, self.env(self))
    if not isinstance(s, schema.Schema):
      raise ValueError('Node %r with schema node %r should evaluate to Schema, got %r' % (member_node, member_node.member_schema, s))
    return s

  def get_required_fields(self):
    """Return the names of fields that are required according to the schema."""
    return [m.name for m in self.__tuplenode.members if m.member_schema.required]

  def _keys_and_privacy(self):
    return {k: self.__tuplenode.member[k].member_schema.private for k in self.keys()}

  def exportable_keys(self):
    """Return a list of keys that are exportable from this tuple."""
    return [k for k, v in self._keys_and_privacy().items() if not v]

  def __iter__(self):
    return iter(self.keys())

  def __repr__(self):
    return '{%s}' % '; '.join(self._render(k) for k in self.keys())

  def __call__(self, right):
    """Apply a tuple to another value."""
    if framework.is_tuple(right):
      return self.compose(right)

    if framework.is_str(right):
      return self[right]

    raise exceptions.EvaluationError("Can't apply tuple (%r) to argument (%r): string or tuple expected" % (self, right))


class CompositeBaseTuple(object):
  """A tuple-like object that will be used to resolve 'base' to.

  This will start looking in the tuples of the composite, from right to left,
  and check the complete composite for declared v
  """
  def __init__(self, composite, index):
    self.composite = composite
    self.index = index

  def __getitem__(self, key):
    for tup, env in self.composite.lookups[self.index:]:
      if key in tup:
        # Can't simply use tup[key] because we need to get the indicated thunk but evaluate it in
        # the CURRENT environment. The environment there is a combination of all tuples to the left
        # of the indicated tuple.
        thunk = tup.get_thunk(key)
        if not isinstance(thunk, framework.Thunk):
          return thunk
        if not thunk.is_unbound():
          return framework.eval(thunk, env)
    raise exceptions.EvaluationError('Unknown key: base.%r in composite tuple:' % (key, self.composite))


def env_of(tup, self):
  if isinstance(tup, Tuple):
    return tup.env(self)
  return tup


class CompositeTuple(Tuple):
  """2 or more composited tuples.

  Keys are looked up from right-to-left, and every key will be evaluated in its
  tuple's own environment, except the 'current_scope' will be set to the
  CompositeTuple (so that declared names will be looked up in the composite
  tuple).

  To properly resolve the special variable 'base', we construct smaller
  composite tuples which only contain the tuples to the left of each tuple,
  which will get returned as the result of the expression 'base'.
  """
  def __init__(self, tuples, dict2tuple):
    self.ident = framework.obj_ident()
    self._tuples = tuples
    self._keys = functools.reduce(lambda s, t: s.union(t.keys()), self._tuples, set())
    self._makeLookupList()
    self.tuple_schema = schema.AnySchema()
    self.dict2tuple = dict2tuple

  def _makeLookupList(self):
    # Count index from the back because we're going to reverse
    envs = [framework.Environment({'base': CompositeBaseTuple(self, len(self.tuples) - i)}, env_of(t, self)) for i, t in enumerate(self.tuples)]
    self.lookups = list(zip(self._tuples, envs))
    self.lookups.reverse()

  @property
  def tuples(self):
    return self._tuples

  def __contains__(self, key):
    return key in self._keys

  def keys(self):
    return list(self._keys)

  def items(self):
    return [(k, self[k]) for k in self.keys()]

  def get(self, key, default=None):
    if key in self:
      return self[key]
    return default

  def is_bound(self, name):
    return name in self and self.maybe_get_thunk_env(name)[0] is not None

  def maybe_get_thunk_env(self, key):
    for tup, env in self.lookups:
      if key in tup:
        thunk = tup.get_thunk(key)
        if not isinstance(thunk, framework.Thunk) or not thunk.is_unbound():
          return (thunk, env)
    return None, None

  def get_thunk_env(self, key):
    thunk, env = self.maybe_get_thunk_env(key)
    if thunk:
      return thunk, env
    raise exceptions.EvaluationError('Unknown key: %r in composite tuple %r' % (key, self))

  def exportable_keys(self):
    """Return a list of keys that are exportable from this tuple.

    Returns all keys that are not private in any of the tuples.
    """
    keys = collections.defaultdict(list)
    for tup in self._tuples:
      for key, private in tup._keys_and_privacy().items():
        keys[key].append(private)
    return [k for k, ps in keys.items() if not any(ps)]

  def __repr__(self):
    return ' '.join(repr(t) for t in self.tuples)


#----------------------------------------------------------------------

class OnDiskFiles(object):
  """Abstraction of a file system, with search path."""
  def __init__(self, search_path=[]):
    self.search_path = search_path

  def resolve(self, current_file, rel_path):
    """Search the filesystem."""
    search_path = [path.dirname(current_file)] + self.search_path

    target_path = None
    for search in search_path:
      if path.isfile(path.join(search, rel_path)):
        target_path = path.normpath(path.join(search, rel_path))
        break

    if not target_path:
      raise exceptions.EvaluationError('No such file: %r, searched %s' %
                            (rel_path, ':'.join(search_path)))

    return target_path, path.abspath(target_path)

  def load(self, path):
    with open(path, 'r') as f:
      return f.read()


class InMemoryFiles(object):
  """Simulate a filesystem from an in-memory dictionary.

  The dictionary maps path to file contents.
  """
  def __init__(self, file_dict):
    self.file_dict = file_dict

  def resolve(self, current_file, rel_path):
    """Search the filesystem."""
    p = path.join(path.dirname(current_file), rel_path)
    if p not in self.file_dict:
      raise RuntimeError('No such fake file: %r' % p)
    return p, p

  def load(self, path):
    return self.file_dict[path]





"""
GCL -- Generic Configuration Language

See README.md for an explanation of GCL and concepts.
"""
from os import path
import json

from . import functions
from . import exceptions
from . import schema
from . import ast
from . import runtime
from . import util
from . import framework


__version__ = '0.6.5'


# Namespace copy for backwards compatibility.
GCLError = exceptions.GCLError
ParseError = exceptions.ParseError
EvaluationError = exceptions.EvaluationError

is_tuple = framework.is_tuple
is_list = framework.is_list

TupleLike = framework.TupleLike
Environment = framework.Environment
find_relative = util.find_relative
Tuple = runtime.Tuple

InMemoryFiles = runtime.InMemoryFiles


class NormalLoader(object):
  def __init__(self, fs):
    self.fs = fs
    self.cache = framework.Cache()

  def __call__(self, current_file, rel_path, env=None):
    nice_path, full_path = self.fs.resolve(current_file, rel_path)

    # Cache on full path, but tell script about nice path
    if path.splitext(nice_path)[1] == '.json':
      # Load as JSON
      do_load = lambda: json.loads(self.fs.load(full_path))
    else:
      # Load as GCL
      do_load = lambda: loads(self.fs.load(full_path), filename=nice_path, loader=self, env=env)
    return self.cache.get(full_path, do_load)


def loader_with_search_path(search_path):
  return NormalLoader(runtime.OnDiskFiles(search_path))


make_env = framework.make_env
make_tuple = ast.make_tuple


# Default loader doesn't have any search path
default_loader = NormalLoader(runtime.OnDiskFiles())

#----------------------------------------------------------------------
#  Top-level functions
#

_default_bindings = {}
_default_bindings.update(functions.builtin_functions)
_default_bindings.update(schema.builtin_schemas)

default_env = framework.Environment(_default_bindings)

def reads(s, filename=None, loader=None, implicit_tuple=True):
  """Load but don't evaluate a GCL expression from a string."""
  return ast.reads(s,
      filename or '<input>',
      loader or default_loader,
      implicit_tuple)


def read(filename, loader=None, implicit_tuple=True):
  """Load but don't evaluate a GCL expression from a file."""
  with open(filename, 'r') as f:
    return reads(f.read(),
                 filename=filename,
                 loader=loader,
                 implicit_tuple=implicit_tuple)


mod_schema = schema  # Just because I want to use schema as a keyword argument below
def loads(s, filename=None, loader=None, implicit_tuple=True, env={}, schema=None):
  """Load and evaluate a GCL expression from a string."""
  ast = reads(s, filename=filename, loader=loader, implicit_tuple=implicit_tuple)
  if not isinstance(env, framework.Environment):
    # For backwards compatibility we accept an Environment object. Otherwise assume it's a dict
    # whose bindings will add/overwrite the default bindings.
    env = framework.Environment(dict(_default_bindings, **env))
  obj = framework.eval(ast, env)
  return mod_schema.validate(obj, schema)


def load(filename, loader=None, implicit_tuple=True, env={}, schema=None):
  """Load and evaluate a GCL expression from a file."""
  with open(filename, 'r') as f:
    return loads(f.read(),
                 filename=filename,
                 loader=loader,
                 implicit_tuple=implicit_tuple,
                 env=env,
                 schema=schema)



"""
AST and parsing related functions.
"""
import itertools
import sys

import pyparsing as p

from . import runtime
from . import exceptions
from . import schema
from . import functions
from . import framework


def do(*fns):
  def fg(args):
    for fn in fns:
      args = fn(args)
    return args
  return fg


def doapply(what):
  def fn(args):
    return what(*args)
  return fn


def head(x): return x[0]
def second(x): return x[1]
def inner(x): return x[1:-1]
def mkBool(s): return True if s == 'true' else False
def drop(x): return []


def pafac(fn):
  """Make a function that accepts a parsed string and a SourceLocation into a function that can be
  passed into a setParseAction.
  """
  def wrapped(s, loc, x):
    try:
      return fn(SourceLocation(s, loc), *list(x))
    except TypeError as e:
      # pyparsing will catch TypeErrors to "detect" our arity, but I don't want to swallow errors
      # here. Convert to some other exception type. I'd LOVE to keep the stack trace here, but there
      # is no one syntax that is not a syntax error in either Python 2 or Python 3. So in Python 2
      # we don't keep the stack trace, unfortunately.

      if hasattr(e, 'with_traceback'):
        # Python 3
        raise RuntimeError(str(e)).with_traceback(sys.exc_info()[2])
      else:
        # Python 2, put the original trace inside the error message. This exception is never
        # supposed to happen anyway, it is just for debugging.
        raise RuntimeError(traceback.format_exc())
  return wrapped


class ParseContext(object):
  def __init__(self):
    self.filename = '<from string>'
    self.loader = None

the_context = ParseContext()


class SourceLocation(object):
  def __init__(self, string, offset):
    self.filename = the_context.filename
    self.string = string
    self.offset = offset

  @property
  def line(self):
    return p.line(self.offset, self.string)

  @property
  def lineno(self):
    return p.lineno(self.offset, self.string)

  @property
  def col(self):
    return p.col(self.offset, self.string)

  @property
  def line_spec(self):
    return '%s:%s' % (self.filename, self.lineno)

  def error_in_context(self, msg):
    msg = '%s:%d: %s in \'%s\'' % (self.filename, self.lineno, msg, self.line)
    return msg

  def __str__(self):
    return self.string[:self.offset] + '|' + self.string[self.offset:]

  @staticmethod
  def empty():
    return SourceLocation('', 0)


class Null(framework.Thunk):
  """Null, evaluates to None."""
  def __init__(self):
    self.ident = framework.obj_ident()

  def eval(self, env):
    return None

  def __repr__(self):
    return "null";


class Inherit(framework.BindableThunk):
  """Inherit Thunks can be either bound or unbound."""

  def __init__(self, name, env=None):
    self.ident = framework.obj_ident()
    self.name = name
    self.env = env

  def bind(self, env):
    return Inherit(self.name, env)

  def eval(self, env):
    if not self.env:
      raise exceptions.EvaluationError("Shouldn't evaluate unbound Inherit nodes")
    return self.env[self.name]

  def __repr__(self):
    return 'inherit %s' % self.name


def mkInherits(tokens):
  return [TupleMemberNode(SourceLocation('', 0), t, no_schema, Inherit(t)) for t in list(tokens)]


class Literal(framework.Thunk):
  """A GCL literal expression."""
  def __init__(self, value):
    self.ident = framework.obj_ident()
    self.value = value

  def eval(self, env):
    return self.value

  def __repr__(self):
    if type(self.value) == bool:
      return 'true' if self.value else 'false'
    return repr(self.value)


class Var(framework.Thunk):
  """Reference to another value."""
  def __init__(self, name, location):
    self.ident = framework.obj_ident()
    self.name = name
    self.location = location

  def eval(self, env):
    try:
      return env[self.name]
    except exceptions.EvaluationError as e:
      raise exceptions.EvaluationError(self.location.error_in_context('while evaluating %r' % self.name), e)

  def __repr__(self):
    return self.name


def mkVar(s, loc, toks):
  return Var(toks[0], SourceLocation(s, loc))


class List(framework.Thunk):
  """A GCL list."""
  def __init__(self, values):
    self.ident = framework.obj_ident()
    self.values = list(values)

  def eval(self, env):
    return [framework.eval(v, env) for v in self.values]

  def __repr__(self):
    return repr(self.values)


class ArgList(framework.Thunk):
  """A paren-separated argument list.

  This is actually a shallow wrapper for Python's list type. We can't use that
  because pyparsing will automatically concatenate lists, which we don't want
  in this case.
  """
  def __init__(self, values):
    self.ident = framework.obj_ident()
    self.values = values

  def eval(self, env):
    return [framework.eval(v, env) for v in self.values]

  def __repr__(self):
    return '(%s)' % ', '.join(repr(x) for x in self.values)


class TupleMemberNode(object):
  """AST node for tuple members.

  They have a name, an expression value and an optional schema.
  """
  def __init__(self, sloc, name, schema, value=None):
    self.sloc = sloc
    self.name = name
    self.value = value
    self.member_schema = schema

  def __repr__(self):
    schema_repr = ': %r' % self.member_schema if not isinstance(self.member_schema, NoSchemaNode) else ''
    return '%s%s = %r' % (self.name, schema_repr, self.value)


class TupleNode(framework.Thunk):
  """AST node for tuple

  When evaluating, the tuple doesn't actually evaluate its children. Instead, we return a (lazy)
  Tuple object that only evaluates the elements when they're requested.
  """
  def __init__(self, sloc, *members):
    duplicates = [name for name, ns in itertools.groupby(sorted(m.name for m in members)) if len(list(ns)) > 1]
    if duplicates:
      raise exceptions.ParseError('Key %s occurs more than once in tuple at %s' % (', '.join(duplicates), sloc.error_in_context('')))

    self.ident = framework.obj_ident()
    self.members = members
    self.member = {m.name: m for m in self.members}
    self._cache = framework.Cache()

  def eval(self, env):
    return self._cache.get(env.ident, self._make_tuple, env)

  def _make_tuple(self, env):
    """Instantiate the Tuple based on this TupleNode."""
    t = runtime.Tuple(self, env, dict2tuple)
    # A tuple also provides its own schema spec
    schema = schema_spec_from_tuple(t)
    t.attach_schema(schema)
    return t

  def __repr__(self):
    return ('{' +
            '; '.join(repr(m) for m in self.members) +
            '}')


class RuntimeTupleNode(TupleNode):
  """Fake AST node.

  Tuples require a reference to a TupleNode-like object. However, sometimes we want to convert
  dictionaries to Tuple objects at runtime, so we need to invent a tuple-like object.
  """
  def __init__(self, dct):
    self.members = [TupleMemberNode(SourceLocation.empty(), key, schema=NoSchemaNode(), value=value) for key, value in dct.items()]
    self.member = {m.name: m for m in self.members}


def dict2tuple(dct):
    return runtime.Tuple(RuntimeTupleNode(dct), framework.EmptyEnvironment(), dict2tuple)


class Application(framework.Thunk):
  """Function application."""
  def __init__(self, location, left, right):
    self.location = location
    self.ident = framework.obj_ident()
    self.left = left
    self.right = right

  def right_as_list(self):
    return self.right if isinstance(self.right, ArgList) else ArgList([self.right])

  def eval_right_as_list(self, env):
    return framework.eval(self.right_as_list(), env)

  def eval(self, env):
    fn = framework.eval(self.left, env)

    try:
      # Tuple application
      if isinstance(fn, framework.TupleLike):
        return self.applyTuple(fn, self.eval_right_as_list(env), env)

      # List application
      if isinstance(fn, list) or framework.is_str(fn):
        return self.applyIndex(fn, self.eval_right_as_list(env))

      # Any other callable type, just use as a Python function
      if not callable(fn):
        raise exceptions.EvaluationError('Result of %r (%r) not callable' % (self.left, fn))

      return call_fn(fn, self.right_as_list(), env)
    except Exception as e:
      # Wrap exceptions
      raise exceptions.EvaluationError(self.location.error_in_context('while calling \'%r\'' % self), e)

  def __repr__(self):
    return '%r(%r)' % (self.left, self.right)

  def applyTuple(self, tuple, right, env):
    """Apply a tuple to something else."""
    if len(right) != 1:
      raise exceptions.EvaluationError('Tuple (%r) can only be applied to one argument, got %r' % (self.left, self.right))
    right = right[0]

    return tuple(right)

  def applyIndex(self, lst, right):
    """Apply a list to something else."""
    if len(right) != 1:
      raise exceptions.EvaluationError('%r can only be applied to one argument, got %r' % (self.left, self.right))
    right = right[0]

    if isinstance(right, int):
      return lst[right]

    raise exceptions.EvaluationError("Can't apply %r to argument (%r): integer expected, got %r" % (self.left, self.right, right))


def mkApplications(s, loc, atoms):
  """Make a sequence of applications from a list of tokens.

  atoms is a list of atoms, which will be handled left-associatively. E.g:

      ['foo', [], []] == foo()() ==> Application(Application('foo', []), [])
  """
  location = SourceLocation(s, loc)
  atoms = list(atoms)
  while len(atoms) > 1:
    atoms[0:2] = [Application(location, atoms[0], atoms[1])]

  # Nothing left to apply
  return atoms[0]


class UnOp(framework.Thunk):
  def __init__(self, op, right):
    self.ident = framework.obj_ident()
    self.op = op
    self.right = right

  def eval(self, env):
    fn = functions.unary_operators.get(self.op, None)
    if fn is None:
      raise exceptions.EvaluationError('Unknown unary operator: %s' % self.op)
    return call_fn(fn, ArgList([self.right]), env)

  def __repr__(self):
    return '%s%r' % (self.op, self.right)


def mkUnOp(tokens):
  return UnOp(tokens[0], tokens[1])


class BinOp(framework.Thunk):
  def __init__(self, left, op, right):
    self.ident = framework.obj_ident()
    self.left = left
    self.op = op
    self.right = right

  def eval(self, env):
    fn = functions.all_binary_operators.get(self.op, None)
    if fn is None:
      raise exceptions.EvaluationError('Unknown operator: %s' % self.op)

    return call_fn(fn, ArgList([self.left, self.right]), env)

  def __repr__(self):
    return ('%r %s %r' % (self.left, self.op, self.right))


def mkBinOps(tokens):
  tokens = list(tokens)
  while len(tokens) > 1:
    assert(len(tokens) >= 3)
    tokens[0:3] = [BinOp(tokens[0], tokens[1], tokens[2])]
  return tokens[0]


def call_fn(fn, arglist, env):
  """Call a function, respecting all the various types of functions that exist."""
  if isinstance(fn, framework.LazyFunction):
    # The following looks complicated, but this is necessary because you can't
    # construct closures over the loop variable directly.
    thunks = [(lambda thunk: lambda: framework.eval(thunk, env))(th) for th in arglist.values]
    return fn(*thunks)

  evaled_args = framework.eval(arglist, env)
  if isinstance(fn, framework.EnvironmentFunction):
    return fn(*evaled_args, env=env)

  return fn(*evaled_args)


class Deref(framework.Thunk):
  """Dereferencing of a dictionary-like object."""
  def __init__(self, haystack, needle, location):
    self.ident = framework.obj_ident()
    self.haystack = haystack
    self.needle = needle
    self.location = location

  def eval(self, env):
    try:
      haystack = framework.eval(self.haystack, env)
      return haystack[self.needle]
    except exceptions.EvaluationError as e:
      raise exceptions.EvaluationError(self.location.error_in_context('while evaluating \'%r\'' % self), e)
    except TypeError as e:
      raise exceptions.EvaluationError(self.location.error_in_context('while getting %r from %r' % (self.needle, self.haystack)), e)

  def __repr__(self):
    return '%s.%s' % (self.haystack, self.needle)


def mkDerefs(s, loc, tokens):
  location = SourceLocation(s, loc)
  tokens = list(tokens)
  while len(tokens) > 1:
    tokens[0:2] = [Deref(tokens[0], tokens[1], location)]
  return tokens[0]


class Condition(framework.Thunk):
  def __init__(self, cond, then, else_):
    self.ident = framework.obj_ident()
    self.cond = cond
    self.then = then
    self.else_ = else_

  def eval(self, env):
    if framework.eval(self.cond, env):
      return framework.eval(self.then, env)
    else:
      return framework.eval(self.else_, env)

  def __repr__(self):
    return 'if %r then %r else %r' % (self.cond, self.then, self.else_)


class ListComprehension(framework.Thunk):
  def __init__(self, expr, var, collection, cond=None):
    self.ident = framework.obj_ident()
    self.expr = expr
    self.var = var
    self.collection = collection
    self.cond = cond

  def eval(self, env):
    ret = []
    for x in framework.eval(self.collection, env):
      new_env = framework.Environment({self.var.name:x}, env)
      if self.cond is None or framework.eval(self.cond, new_env):
        ret.append(framework.eval(self.expr, new_env))
    return ret

  def __repr__(self):
    return '[%r for %r in %r]' % (self.expr, self.var, self.collection)


class Void(framework.Thunk):
  """A missing value."""
  def __init__(self, name, location):
    self.name = name
    self.location = location
    self.ident = framework.obj_ident()

  def eval(self, env):
    raise exceptions.EvaluationError(self.location.error_in_context('Unbound value: %r' % self.name))

  def is_unbound(self):
    return True

  def __repr__(self):
    return '<unbound>'


class Include(framework.Thunk):
  def __init__(self, file_ref):
    self.ident = framework.obj_ident()
    self.file_ref = file_ref
    self.current_file = the_context.filename
    self.loader = the_context.loader

  def eval(self, env):
    file_ref = framework.eval(self.file_ref, env)
    if not framework.is_str(file_ref):
      raise exceptions.EvaluationError('Included argument (%r) must be a string, got %r' %
                            (self.file_ref, file_ref))

    return self.loader(self.current_file, file_ref, env=env.root)

  def __repr__(self):
    return 'include(%r)' % self.file_ref

#----------------------------------------------------------------------
#  Schema AST model
#

class NoSchemaNode(framework.Thunk):
  """For values without a schema."""
  def __init__(self):
    self.ident = framework.obj_ident()
    self.required = False
    self.private = False

  def eval(self, env):
    return schema.AnySchema()

no_schema = NoSchemaNode()  # Singleton object


class AnySchemaExprNode(framework.Thunk):
  def __init__(self):
    self.ident = framework.obj_ident()

  def eval(self, env):
    return schema.AnySchema()

any_scheam_expr = AnySchemaExprNode()


class MemberSchemaNode(framework.Thunk):
  """AST node for member schema definitions. Can be evaluated to produce runtime Schema classes.

  NOTE: this class does a little funky logic. Because we want to be able to write:

      var : int;

  Instead of 'var : "int"', some words are special-cased. However, we also want to be able to parse
  things of the form:

      tuple_with_int_foo : { foo : int };
      tuple_of_type : SomeObject;

  Which requires us to parse the grammar 'expression+schema_literals' (as opposed to just the
  grammar 'expression'), which I don't really want to double-define. Instead, we'll just pretend
  certain variable names like 'int' and 'string' refer to global objects that can't be rebound (in
  effect, they resolve to their variable name).

  The schema evaluation object model looks like this:


      TupleNode <>------> TupleMemberNode <>---------> MemberSchemaNode
         |                                                   |
         | eval()                                            | eval()
         v                                                   v
      Tuple                                           schema.Schema() +validate()
                                                             /\
                                                        TupleSchema


  """
  def __init__(self, sloc, private, required, expr):
    self.sloc = sloc
    self.private = private
    self.required = required
    self.expr = expr
    self.ident = framework.obj_ident()

  def eval(self, env):
    return make_schema_from(self.expr, env)

  def __repr__(self):
    return ' '.join((['required'] if self.required else []) +
                    ([repr(self.expr)] if not isinstance(self.expr, NoSchemaNode) else []))


class TupleSchemaAccess(object):
  """A class that behaves like a dictionary and returns member schemas from a tuple."""
  def __init__(self, tuple):
    self.tuple = tuple

  def __getitem__(self, key):
    return self.tuple.get_schema_spec(key)

  def __contains__(self, key):
    return key in self.tuple

  def get(self, key, default=None):
    return self[key] if key in self else default

  def values(self):
    return [self[k] for k in self.tuple.keys()]

  def __repr__(self):
    return 'TupleSchemaAccess(%r)' % self.tuple


def schema_spec_from_tuple(tup):
  """Return the schema spec from a run-time tuple."""
  if hasattr(tup, 'get_schema_spec'):
    # Tuples have a TupleSchema field that contains a model of the schema
    return schema.from_spec({
        'fields': TupleSchemaAccess(tup),
        'required': tup.get_required_fields()})
  return schema.AnySchema()


def make_schema_from(value, env):
  """Make a Schema object from the given spec.

  The input and output types of this function are super unclear, and are held together by ponies,
  wishes, duct tape, and a load of tests. See the comments for horrific entertainment.
  """

  # So this thing may not need to evaluate anything[0]
  if isinstance(value, framework.Thunk):
    value = framework.eval(value, env)

  # We're a bit messy. In general, this has evaluated to a Schema object, but not necessarily:
  # for tuples and lists, we still need to treat the objects as specs.
  if isinstance(value, schema.Schema):
    return value

  if framework.is_tuple(value):
    # If it so happens that the thing is a tuple, we need to pass in the data in a bit of a
    # different way into the schema factory (in a dictionary with {fields, required} keys).
    return schema_spec_from_tuple(value)

  if framework.is_list(value):
    # [0] This list may contain tuples, which oughta be treated as specs, or already-resolved schema
    # objects (as returned by 'int' and 'string' literals). make_schema_from
    # deals with both.
    return schema.from_spec([make_schema_from(x, env) for x in value])

  raise exceptions.EvaluationError('Can\'t make a schema from %r' % value)


def make_tuple(x):
  """Turn a dict-like object into a Tuple."""
  if isinstance(x, framework.TupleLike):
    return x
  return dict2tuple(x)

#----------------------------------------------------------------------
#  Grammar
#

def sym(sym):
  return p.Suppress(sym)


def kw(kw):
  return p.Keyword(kw).suppress()


def listMembers(sep, expr, what):
  return p.Optional(p.delimitedList(expr, sep) -
                    p.Optional(sep).suppress()).setParseAction(what)


def bracketedList(l, r, sep, expr, what):
  """Parse bracketed list.

  Empty list is possible, as is a trailing separator.
  """
  return (sym(l) - listMembers(sep, expr, what) - sym(r)).setParseAction(head)


keywords = ['and', 'or', 'not', 'if', 'then', 'else', 'include', 'inherit', 'null', 'true', 'false',
    'for', 'in']

expression = p.Forward()

comment = '#' + p.restOfLine

quotedIdentifier = p.QuotedString('`', multiline=False)

# - Must start with an alphascore
# - May contain alphanumericscores and special characters such as : and -
# - Must not end in a special character
identifier = quotedIdentifier | p.Regex(r'[a-zA-Z_]([a-zA-Z0-9_:-]*[a-zA-Z0-9_])?')

# Contants
integer = p.Word(p.nums).setParseAction(do(head, int, Literal))
floating = p.Regex(r'\d*\.\d+').setParseAction(do(head, float, Literal))
dq_string = p.QuotedString('"', escChar='\\', multiline=True).setParseAction(do(head, Literal))
sq_string = p.QuotedString("'", escChar='\\', multiline=True).setParseAction(do(head, Literal))
boolean = (p.Keyword('true') | p.Keyword('false')).setParseAction(do(head, mkBool, Literal))
null = p.Keyword('null').setParseAction(Null)

# List
list_ = bracketedList('[', ']', ',', expression, List)

# Tuple
inherit = (kw('inherit') - p.ZeroOrMore(identifier)).setParseAction(mkInherits)
schema_spec = (p.Optional(p.Keyword('private').setParseAction(lambda: True), default=False)
               - p.Optional(p.Keyword('required').setParseAction(lambda: True), default=False)
               - p.Optional(expression, default=any_scheam_expr)).setParseAction(pafac(MemberSchemaNode))
optional_schema = p.Optional(p.Suppress(':') - schema_spec, default=no_schema)
tuple_member = (inherit
               | (identifier + optional_schema + ~p.FollowedBy('=')).setParseAction(lambda s, loc, x: TupleMemberNode(SourceLocation(s, loc), x[0], x[1], Void(x[0], SourceLocation(s, loc))))
               | (identifier - optional_schema - p.Suppress('=') - expression).setParseAction(pafac(TupleMemberNode))
               )
tuple_members = listMembers(';', tuple_member, pafac(TupleNode))
tuple = bracketedList('{', '}', ';', tuple_member, pafac(TupleNode))

# Variable (can't be any of the keywords, which may have lower matching priority)
variable = ~p.Or([p.Keyword(k) for k in keywords]) + identifier.copy().setParseAction(mkVar)

# Argument list will live by itself as a atom. Actually, it's a tuple, but we
# don't call it that because we use that term for something else already :)
arg_list = bracketedList('(', ')', ',', expression, ArgList)

parenthesized_expr = (sym('(') - expression - ')').setParseAction(head)

unary_op = (p.oneOf(' '.join(functions.unary_operators.keys())) - expression).setParseAction(mkUnOp)

if_then_else = (kw('if') + expression +
                kw('then') - expression -
                kw('else') - expression).setParseAction(doapply(Condition))

list_comprehension = (sym('[') + expression + kw('for') - variable - kw('in') -
    expression - p.Optional(kw('if') - expression) - sym(']')).setParseAction(doapply(ListComprehension))


# We don't allow space-application here
# Now our grammar is becoming very dirty and hackish
deref = p.Forward()
include = (kw('include') - deref).setParseAction(doapply(Include))

atom = (tuple
        | variable
        | dq_string
        | sq_string
        | boolean
        | list_comprehension
        | list_
        | null
        | unary_op
        | parenthesized_expr
        | if_then_else
        | include
        | floating
        | integer
        )

# We have two different forms of function application, so they can have 2
# different precedences. This one: fn(args), which binds stronger than
# dereferencing (fn(args).attr == (fn(args)).attr)
applic1 = (atom - p.ZeroOrMore(arg_list)).setParseAction(mkApplications)

# Dereferencing of an expression (obj.bar)
deref << (applic1 - p.ZeroOrMore(p.Suppress('.') - identifier)).setParseAction(mkDerefs)

# All binary operators at various precedence levels go here:
# This piece of code does the moral equivalent of:
#
#     T = F*F | F/F | F
#     E = T+T | T-T | T
#
# etc.
term = deref
for op_level in functions.binary_operators_before_juxtaposition:
  operator_syms = ' '.join(op_level.keys())
  term = (term - p.ZeroOrMore(p.oneOf(operator_syms) - term)).setParseAction(mkBinOps)

# Juxtaposition function application (fn arg), must be 1-arg every time
applic2 = (term - p.ZeroOrMore(term)).setParseAction(mkApplications)

term = applic2
for op_level in functions.binary_operators_after_juxtaposition:
  operator_syms = ' '.join(op_level.keys())
  term = (term - p.ZeroOrMore(p.oneOf(operator_syms) - term)).setParseAction(mkBinOps)

expression << term

# Two entry points: start at an arbitrary expression, or expect the top-level
# scope to be a tuple.
start = expression.ignore(comment)
start_tuple = tuple_members.ignore(comment)


def reads(s, filename, loader, implicit_tuple):
  """Load but don't evaluate a GCL expression from a string."""
  try:
    the_context.filename = filename
    the_context.loader = loader
    return (start_tuple if implicit_tuple else start).parseWithTabs().parseString(s, parseAll=True)[0]
  except (p.ParseException, p.ParseSyntaxException) as e:
    msg = '%s:%d: %s\n%s\n%s^-- here' % (the_context.filename, e.lineno, e.msg, e.line, ' ' * (e.col - 1))
    raise exceptions.ParseError(msg)




from . import exceptions

class Thunk(object):
  """An object that is evaluatable in a scope."""
  def eval(self, env):
    raise exceptions.EvaluationError('Not implemented')

  def is_unbound(self):
    """Return whether this thunk represents an actual value (not a placeholder for an unbound value)."""
    return False


class BindableThunk(Thunk):
  """An thunk that is supposed to be bound to a scope before being evaluated.

  We only have objects that are bindable to a tuple's parent environment right now.

  The result of the bind operation should still be a thunk.
  """
  def bind(self, parent_env):
    raise exceptions.EvaluationError('Not implemented')


def eval(thunk, env):
  """Evaluate a thunk in an environment.

  Will defer the actual evaluation to the thunk itself, but adds two things:
  caching and recursion detection.

  Since we have to use a global evaluation stack (because there is a variety of functions that may
  be invoked, not just eval() but also __getitem__, and not all of them can pass along a context
  object), GCL evaluation is not thread safe.

  With regard to schemas:

  - A schema can be passed in from outside. The returned object will be validated to see that it
    conforms to the schema. The schema will be attached to the value if possible.
  - Some objects may contain their own schema, such as tuples. This would be out of scope of the
    eval() function, were it not for:
  - Schema validation can be disabled in an evaluation call stack. This is useful if we're
    evaluating a tuple only for its schema information. At that point, we're not interested if the
    object is value-complete.
  """
  key = Activation.key(thunk, env)
  if Activation.activated(key):
    raise exceptions.RecursionError('Reference cycle')

  with Activation(key):
    return eval_cache.get(key, thunk.eval, env)


class TupleLike(object):
  """Interface for tuple-like objects."""
  def __getitem__(self, key):
    pass

  def keys(self):
    pass

  def items(self):
    pass

  def __contains__(self):
    pass

  def __iter__(self):
    pass

  def is_bound(self, name):
    """Return whether the value exists and is bound."""
    pass

  def has_key(self, name):
    return name in self

  def exportable_keys(self):
    """Return a list of keys that are exportable from this tuple."""
    pass


# Because we can't trust id(), it'll get reused, we number objects ourselves
# for caching purposes.
obj_nr = 0

def obj_ident():
  global obj_nr
  obj_nr += 1
  return obj_nr


class Activation(object):
  """Activation tracker.

  An activation set will contain all the evals() that are activated at any point in time, to detect
  infinite recursion during evaluation.

  This class has global data because we can't really pass references through all the various ways
  that we can start evaluating GCL constructs (notably dict element accessors), so it must be
  out-of-band in a global variable. This is the reason why GCL evaluation is NOT thread safe.
  """
  activation_set = {}

  def __init__(self, key):
    self.key = key

  def __enter__(self):
    self.activation_set[self.key] = self

  def __exit__(self, value, type, exc):
    del self.activation_set[self.key]

  @classmethod
  def activated(cls, key):
    return key in cls.activation_set

  @classmethod
  def key(cls, thunk, env):
    return (thunk.ident, env.ident)


class Environment(object):
  """Binding environment, inherits from another Environment."""

  def __init__(self, values, parent=None, names=None):
    self.ident = obj_ident()
    self.parent = parent or EmptyEnvironment()
    self.values = values
    self.names = names or values.keys()

  def __getitem__(self, key):
    if key in self.names:
      getter = getattr(self.values, 'get_no_validate', None)
      if getter:
        return getter(key)
      return self.values[key]
    return self.parent[key]

  def __contains__(self, key):
    if key in self.names:
      return True
    return key in self.parent

  @property
  def root(self):
    if isinstance(self.parent, EmptyEnvironment):
      return self
    else:
      return self.parent.root

  def extend(self, d):
    return Environment(d or {}, self)

  def keys(self):
    return set(self.names).union(self.parent.keys())

  def __repr__(self):
    return 'Environment(%s :: %r)' % (', '.join(self.names), self.parent)


def make_env(x):
  """Turn an dict into an Environment object."""
  return x if isinstance(x, Environment) else Environment(x)


class EmptyEnvironment(object):
  def __init__(self):
    self.ident = obj_ident()

  def __getitem__(self, key):
    raise exceptions.EvaluationError('Unbound variable: %r' % key)

  def __contains__(self, key):
    return False

  def __repr__(self):
    return '<empty>'

  def keys(self):
    return set()

  @property
  def root(self):
    return self

  def extend(self, d):
    return Environment(d or {}, self)


class Cache(object):
  def __init__(self):
    self._cache = {}

  def get(self, key, getter, *args, **kwargs):
    if key not in self._cache:
      self._cache[key] = getter(*args, **kwargs)
    return self._cache[key]


class FunctionMarker(object):
  """Base class for function markers"""
  def __init__(self, fn):
    self.fn = fn

  def __call__(self, *args, **kwargs):
    return self.fn(*args, **kwargs)


class EnvironmentFunction(FunctionMarker):
  """Marker class for a special function that can use the env."""
  pass


class LazyFunction(FunctionMarker):
  """Marker class for functions that take thunks."""
  pass


eval_cache = Cache()

activation_stack = {}


def is_tuple(x):
  return hasattr(x, 'keys')


def is_list(x):
  from . import ast  # I feel dirty
  return isinstance(x, (list, ast.List))


# Python 2 and 3 compatible string check
try:
    isinstance("", basestring)
    def is_str(s):
        return isinstance(s, basestring)
except NameError:
    def is_str(s):
        return isinstance(s, str)





class GCLError(RuntimeError):
  pass

class ParseError(GCLError):
  pass

class EvaluationError(GCLError):
  def __init__(self, message, inner=None):
    super(EvaluationError, self).__init__(message)
    self.inner = inner

  def __str__(self):
    return self.args[0] + ('\n' + str(self.inner) if self.inner else '')

class RecursionError(EvaluationError):
  pass

class SchemaError(EvaluationError):
  pass



import argparse
import json
import sys

import gcl
from gcl import query
from gcl import util


def select(dct, path):
  for part in path:
    if not hasattr(dct, 'keys'):
      raise RuntimeError('Value %r cannot be indexed with %r' % (dct, part))
    if part not in dct:
      raise RuntimeError('Value %r has no key %r' % (dct, part))
    dct = dct[part]
  return dct


def main(argv=None, stdin=None):
  parser = argparse.ArgumentParser(description='Convert (parts of) a GCL model file to JSON.')
  parser.add_argument('file', metavar='FILE', type=str, nargs='?',
                      help='File to parse')
  parser.add_argument('selectors', metavar='SELECTOR', type=str, nargs='*',
                      help='Select nodes to include in the JSON.')
  parser.add_argument('--root', '-r', metavar='PATH', type=str, default='',
                      help='Use the indicated root path as the root of the output JSON object (like a.b.c but without wildcards)')

  args = parser.parse_args(argv or sys.argv[1:])

  try:
    if args.file and args.file != '-':
      model = gcl.load(args.file)
    else:
      model = gcl.loads((stdin or sys.stdin).read(), filename='<stdin>')

    sels = query.GPath(args.selectors)
    if not sels.everything():
      model = sels.select(model).deep()

    plain = util.to_python(model)

    selectors = args.root.split('.') if args.root else []
    selected = select(plain, selectors)

    sys.stdout.write(json.dumps(selected, indent=2))
  except (gcl.ParseError, RuntimeError) as e:
    sys.stderr.write(str(e) + '\n')
    sys.exit(1)
